#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020-2021 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

from argparse import ArgumentParser
from . import entry


__all__ = ['entry']


def get_argparser() -> ArgumentParser:
    parser = ArgumentParser(
        'proto2rapidjson', description='Convert .proto file to header-only RapidJSON based c++ code')
    parser.add_argument('-i', '--input', type=str, dest='input',
                        help='input .proto file', required=True)
    parser.add_argument('-o', '--output', type=str, dest='output',
                        help='output .h file', required=True)
    parser.add_argument('-y', dest='yes', action='store_true',
                        help='pass all interactive checks', default=False)
    return parser


def main():
    args = get_argparser().parse_args()
    entry(args.input, args.output, args.yes)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020-2021 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

import os
from .lexer import scan
from .parser import Parser

__all__ = ['lexer', 'scan', 'parser', 'Parser', 'entry']


def set_comment(s: str) -> str:
    """set the indent of each line in `s` `indent`"""
    lines = s.splitlines(False)
    new_lines = []
    for line in lines:
        line = '// ' + line
        new_lines.append(line)
    return '\n'.join(new_lines)+'\n'


def entry(fin_name: str, fout_name: str, overwrite: bool) -> None:
    with open(fin_name, 'r') as fin:
        s = fin.read()
    if os.path.exists(fout_name) and not overwrite:
        check = input(
            f'File {fout_name} is existed, do you want to overwrite it? [y/N]')
        if check.lower() == 'y' or check == '':
            pass
        else:
            return
    tokens = scan(s)
    parser = Parser(tokens, os.path.splitext(os.path.basename(fin_name))[0])
    with open(fout_name, 'w') as fout:
        fout.write('// ******************************\n')
        fout.write(f'// Origin: {fin_name}\n// Content:\n')
        fout.write(set_comment(s))
        fout.write('// ******************************\n\n')
        fout.write(parser.tocpp())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020-2021 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

from enum import Enum
from typing import List, NamedTuple
try:
    from typing import OrderedDict
    OrderedDictType = OrderedDict
except ImportError:
    from typing import MutableMapping
    from collections import OrderedDict
    OrderedDictType = MutableMapping
from .lexer import TYPE_RESERVED_WORDS, Token, TokenKind

__all__ = ['TokenError', 'ElementKind', 'Element', 'Message', 'Parser']


class TokenError(Exception):
    def __init__(self, token: Token, message: str = 'Unexpected token') -> None:
        self.token = token
        self.message = f'{message} <{token.content}>({token.kind}) in line {token.line}'
        super().__init__(self.message)


API_NAME = {
    'parse_entry': 'FromString',
    'parse_worker': 'FromValue',
    'stringify_entry': 'ToString',
    'stringify_worker': 'ToValue',
    'pretty_stringify_entry': 'ToPrettyString'
}


class ElementKind(Enum):
    custom = 0
    double = 1
    float = 2
    int32 = 3
    uint32 = 4
    int64 = 5
    uint64 = 6
    bool = 7
    string = 8

    def __str__(self):
        d = {
            ElementKind.double: 'double',
            ElementKind.float: 'float',
            ElementKind.int32: 'int32_t',
            ElementKind.uint32: 'uint32_t',
            ElementKind.int64: 'int64_t',
            ElementKind.uint64: 'uint64_t',
            ElementKind.bool: 'bool',
            ElementKind.string: 'std::string'
        }
        return d[self]


def set_indent(s: str, indent: int, newline: bool) -> str:
    """set the indent of each line in `s` `indent`"""
    lines = s.splitlines(False)
    new_lines = []
    for line in lines:
        line = ' '*indent + line
        new_lines.append(line)
    return '\n'.join(new_lines)+('\n' if newline else '')


class Element(NamedTuple):
    identifier: str
    kind: ElementKind
    kindstr: str
    repeated: bool

    def to_declaration(self) -> str:
        if self.repeated:
            return f'std::vector<{self.kindstr}> {self.identifier};\n'
        else:
            return f'{self.kindstr} {self.identifier};\n'

    def to_parse(self) -> str:
        if self.repeated:
            base = 'i'
        else:
            base = f'v["{self.identifier}"]'
        check_function = {
            ElementKind.double: f'{base}.IsDouble()',
            ElementKind.float: f'{base}.IsFloat()',
            ElementKind.int32: f'{base}.IsInt()',
            ElementKind.uint32: f'{base}.IsUint()',
            ElementKind.int64: f'{base}.IsInt64()',
            ElementKind.uint64: f'{base}.IsUint64()',
            ElementKind.bool: f'{base}.IsBool()',
            ElementKind.string: f'{base}.IsString()',
            ElementKind.custom: f'{base}.IsObject()'
        }
        get_function = {
            ElementKind.double: f'{base}.GetDouble()',
            ElementKind.float: f'{base}.GetFloat()',
            ElementKind.int32: f'{base}.GetInt()',
            ElementKind.uint32: f'{base}.GetUint()',
            ElementKind.int64: f'{base}.GetInt64()',
            ElementKind.uint64: f'{base}.GetUint64()',
            ElementKind.bool: f'{base}.GetBool()',
            ElementKind.string: f'{base}.GetString()',
            ElementKind.custom: f'{self.kindstr}().{API_NAME["parse_worker"]}({base})'
        }

        if self.repeated:
            return f'''// parse {self.identifier}
assert(v.HasMember("{self.identifier}"));
assert(v["{self.identifier}"].IsArray());
for (auto&& i : v["{self.identifier}"].GetArray()) {{
    assert({check_function[self.kind]});
    {self.identifier}.emplace_back({get_function[self.kind]});
}}
'''
        else:
            return f'''// parse {self.identifier}
assert(v.HasMember("{self.identifier}"));
assert({check_function[self.kind]});
{self.identifier} = {get_function[self.kind]};
'''

    def to_stringify(self) -> str:
        if self.repeated:
            base = 'i'
        else:
            base = self.identifier
        if self.kind == ElementKind.custom:
            body = f'{base}.ToValue(allocator)'
        else:
            body = f'{base}'
        if self.repeated:
            if self.kind == ElementKind.string:
                # overwrite
                return f'''// stringify {self.identifier}
a.SetArray();
for (auto&& i : {self.identifier}) {{
    if (copy) {{
        a.PushBack(rapidjson::Value().SetString(i.data(), i.size(), allocator), allocator);
    }} else {{
        a.PushBack(rapidjson::StringRef(i), allocator);
    }}
}}
v.AddMember("{self.identifier}", a, allocator);
'''
            else:
                return f'''// stringify {self.identifier}
a.SetArray();
for (auto&& i : {self.identifier}) {{
    a.PushBack({body}, allocator);
}}
v.AddMember("{self.identifier}", a, allocator);
'''
        else:
            return f'''// stringify {self.identifier}
v.AddMember("{self.identifier}", {body}, allocator);
'''


class Message(NamedTuple):
    identifier: str
    elements: OrderedDictType[str, Element]

    def to_struct(self) -> str:
        define_variable = f'{set_indent("".join(e.to_declaration() for e in self.elements.values()), 4, True)}'
        parse_entry = \
            f'''    {self.identifier}& {API_NAME["parse_entry"]}(const char* str) {{
        rapidjson::Document document;
        document.Parse(str);
        assert(document.IsObject());
        return {API_NAME["parse_worker"]}(document);
    }}
'''
        parse_worker = \
            f'''    {self.identifier}& {API_NAME["parse_worker"]}(const rapidjson::Value& v) {{
{set_indent(''.join(e.to_parse() for e in self.elements.values()), 8, False)}
        return *this;
    }}
'''
        stringify_entry = \
            f'''    std::string {API_NAME["stringify_entry"]}(int maxDecimalPlaces = 6) {{
        rapidjson::Document document;
        document.SetObject() = {API_NAME["stringify_worker"]}(document.GetAllocator());
        rapidjson::StringBuffer buffer;
        rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
        writer.SetMaxDecimalPlaces(maxDecimalPlaces);
        document.Accept(writer);
        return buffer.GetString();
    }}
'''

        pretty_stringify_entry = \
            f'''    std::string {API_NAME["pretty_stringify_entry"]}(int maxDecimalPlaces = 6) {{
        rapidjson::Document document;
        document.SetObject() = {API_NAME["stringify_worker"]}(document.GetAllocator());
        rapidjson::StringBuffer buffer;
        rapidjson::PrettyWriter<rapidjson::StringBuffer> writer(buffer);
        writer.SetMaxDecimalPlaces(maxDecimalPlaces);
        document.Accept(writer);
        return buffer.GetString();
    }}
'''

        anylist = any(e.repeated for e in self.elements.values())
        arraystr = 'rapidjson::Value a;\n' if anylist else ''
        stringify_worker = \
            f'''    rapidjson::Value ToValue(rapidjson::Document::AllocatorType& allocator, bool copy = false) {{
        (void)copy;
        rapidjson::Value v(rapidjson::kObjectType);
        {arraystr}
{set_indent(''.join(e.to_stringify() for e in self.elements.values()), 8, False)}
        return v;
    }}
'''

        return f'''struct {self.identifier} {{
{define_variable}
{parse_entry}
{parse_worker}
{stringify_entry}
{pretty_stringify_entry}
{stringify_worker}
}};
'''


class Parser:
    def __init__(self, tokens: List[Token], unique_macro: str) -> None:
        self.tokens = tokens
        self.unique_macro = unique_macro
        self.package: str
        self.messages: OrderedDictType[str, Message] = OrderedDict()
        self.parse()

    def try_match_reserved(self, word: str) -> bool:
        token = self.tokens[0]
        if token.kind == TokenKind.RESERVED and token.content == word:
            self.tokens = self.tokens[1:]
            return True
        else:
            return False

    def match_reserved(self, word: str) -> None:
        token = self.tokens[0]
        if token.kind == TokenKind.RESERVED and token.content == word:
            self.tokens = self.tokens[1:]
        else:
            raise TokenError(token)

    def try_match_eof(self) -> bool:
        token = self.tokens[0]
        if token.kind == TokenKind.EOF:
            self.tokens = self.tokens[1:]
            return True
        else:
            return False

    def match_identifier(self, package: bool = False) -> str:
        token = self.tokens[0]
        if token.kind == TokenKind.IDENTIDIER and (package or '.' not in token.content):
            self.tokens = self.tokens[1:]
            return token.content
        else:
            raise TokenError(token)

    def match_number(self) -> int:
        token = self.tokens[0]
        if token.kind == TokenKind.NUMBER:
            self.tokens = self.tokens[1:]
            return int(token.content)
        else:
            raise TokenError(token)

    def parse_element(self) -> Element:
        repeated = self.try_match_reserved('repeated')
        type_token = self.tokens[0]
        if type_token.kind == TokenKind.IDENTIDIER and type_token.content in self.messages:
            kind = ElementKind.custom
            kindstr = type_token.content
        elif type_token.kind == TokenKind.RESERVED and type_token.content in TYPE_RESERVED_WORDS:
            kind = ElementKind[type_token.content]
            kindstr = str(kind)
        else:
            raise TokenError(type_token, 'Unknown type')
        self.tokens = self.tokens[1:]
        identifier = self.match_identifier()
        # only for compatibility
        if self.try_match_reserved('='):
            self.match_number()
        self.match_reserved(';')
        return Element(identifier, kind, kindstr, repeated)

    def parse_message(self) -> None:
        self.match_reserved('message')
        identifier = self.match_identifier()
        message = Message(identifier, OrderedDict())
        self.match_reserved('{')
        while not self.try_match_reserved('}'):
            element = self.parse_element()
            if element.identifier in message.elements:
                raise KeyError(
                    f'Element {element.identifier} is defined repeatedly')
            else:
                message.elements[element.identifier] = element
        if identifier in self.messages:
            raise KeyError(f'Message {identifier} is defined repeatedly')
        else:
            self.messages[identifier] = message

    def parse_package(self) -> None:
        self.match_reserved('package')
        package = self.match_identifier(package=True)
        self.match_reserved(';')
        if package[0] == '.' or package[-1] == '.':
            raise ValueError(f'package name {package} is invalid')
        if '.' in package:
            print("Attention: nested namespace is used, which is only supported by C++20")
        self.package = package.replace('.', '::')

    def parse(self) -> None:
        self.parse_package()
        self.parse_message()
        while not self.try_match_eof():
            self.parse_message()

    def tocpp(self) -> str:
        return f'''#ifndef _PROTO2RAPIDJSON_{self.unique_macro.upper()}_{self.package.replace('::', '_').upper()}_HEADER_
#define _PROTO2RAPIDJSON_{self.unique_macro.upper()}_{self.package.replace('::', '_').upper()}_HEADER_

// Generated by Proto2RapidJSON
// https://github.com/Sweetnow/proto2rapidjson

#define RAPIDJSON_HAS_STDSTRING 1

#include <string.h>

#include <vector>
#include <string>

#include "rapidjson/document.h"
#include "rapidjson/writer.h"
#include "rapidjson/prettywriter.h"

namespace {self.package} {{
{''.join(m.to_struct() for m in self.messages.values())}
}}  // namespace {self.package}
#endif
'''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020-2021 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

from typing import List, NamedTuple
from enum import Enum

__all__ = ['TYPE_RESERVED_WORDS',
           'RESERVED_WORDS', 'TokenKind', 'Token', 'scan']


TYPE_RESERVED_WORDS = ['double', 'float', 'int32',
                       'uint32', 'int64', 'uint64', 'bool', 'string']
RESERVED_WORDS = ['message', 'package',
                  '{', '}', ';', 'repeated', '//', '='] + TYPE_RESERVED_WORDS


class TokenKind(Enum):
    RESERVED = 0
    IDENTIDIER = 1
    NUMBER = 2
    EOF = 3


class Token(NamedTuple):
    content: str
    kind: TokenKind
    line: int


def scan(input: str) -> List[Token]:
    tokens: List[Token] = []
    for i, line in enumerate(input.splitlines(False)):
        line = line.strip()
        while(len(line) > 0):
            # scan reserved words
            reserved = False
            for word in RESERVED_WORDS:
                if line.startswith(word):
                    if word == '//':
                        line = ''
                    else:
                        tokens.append(Token(word, TokenKind.RESERVED, i+1))
                        line = line[len(word):]
                    reserved = True
                    break
            if reserved:
                line = line.strip()
                continue
            # scan identifier
            identifier = ''
            i = 0
            while i < len(line):
                if line[:i+1].replace('.', '_').isidentifier():
                    i += 1
                    continue
                else:
                    break
            identifier = line[:i]
            if identifier != '':
                line = line[len(identifier):]
                tokens.append(Token(identifier, TokenKind.IDENTIDIER, i+1))
                line = line.strip()
                continue
            # scan number
            number = ''
            for c in line:
                if c.isdigit():
                    number += c
                else:
                    break
            if number != '':
                line = line[len(number):]
                tokens.append(Token(number, TokenKind.NUMBER, i+1))
                number = int(number)
                line = line.strip()
                continue
            # report error
            raise ValueError(f"invalid input: {line}")

    tokens.append(Token('', TokenKind.EOF, -1))
    return tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

import os
import subprocess
from proto2rapidjson import entry


def pipeline(name: str, cpp_relative_path: str = '../cpp'):
    parent = os.path.dirname(__file__)
    cpp_path = os.path.join(parent, cpp_relative_path)
    # Step 1: proto -> h
    proto_path = os.path.join(parent, f'proto/{name}.proto')
    header_path = os.path.join(cpp_path, 'proto.h')
    entry(proto_path, header_path, True)

    # Step 2: h -> out
    subprocess.call(['make', 'clean'], cwd=cpp_path)
    assert(subprocess.call(['make'], cwd=cpp_path) == 0)

    # Step 3: parse json
    exe_path = os.path.join(cpp_path, 'main.out')
    json_path = os.path.join(parent, f'json/{name}.json')
    assert(subprocess.call([exe_path, json_path]) == 0)

    # clean
    subprocess.call(['make', 'clean'], cwd=cpp_path)
    os.remove(header_path)

def test_type():
    pipeline('type')

def test_nest():
    pipeline('nest')

def test_array():
    pipeline('array')

def test_simple():
    pipeline('simple')

def test_more():
    pipeline('more')

def test_comment():
    pipeline('comment')

def test_string_array():
    pipeline('string_array')

def test_namespace():
    pipeline('namespace', '../cpp_namespace')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

import os
import subprocess
from typing import Type
import pytest
from proto2rapidjson import entry
from proto2rapidjson.parser import TokenError


def pipeline(name: str, error: Type[Exception]):
    # Step 1: proto -> h
    parent = os.path.dirname(__file__)
    proto_path = os.path.join(parent, f'proto/{name}.proto')
    header_path = os.path.join(parent, '../cpp/proto.h')
    with pytest.raises(error):
        entry(proto_path, header_path, True)

    if os.path.exists(header_path):
        os.remove(header_path)


def test_nopackage():
    pipeline('nopackage', TokenError)


def test_nomessage():
    pipeline('nomessage', TokenError)


def test_invalid_id():
    pipeline('invalid_id', ValueError)


def test_broken_message():
    pipeline('broken_message', TokenError)


def test_use_before_define():
    pipeline('use_before_define', TokenError)


def test_noid():
    pipeline('noid', TokenError)


def test_nonumber():
    pipeline('nonumber', TokenError)


def test_message_same_id():
    pipeline('message_same_id', Exception)


def test_element_same_id():
    pipeline('element_same_id', Exception)


def test_dot_in_message_name():
    pipeline('dot_in_message_name', TokenError)


def test_dot_in_element_name():
    pipeline('dot_in_element_name', TokenError)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ***********************************************************************************
# * Copyright 2020 Jun Zhang. All Rights Reserved.                                  *
# * Distributed under MIT license.                                                  *
# * See file LICENSE for details for copy at https://opensource.org/licenses/MIT    *
# ***********************************************************************************

import os
import subprocess
from proto2rapidjson import entry


def pipeline(name: str):
    # Step 1: proto -> h
    parent = os.path.dirname(__file__)
    proto_path = os.path.join(parent, f'proto/standard.proto')
    header_path = os.path.join(parent, '../cpp/proto.h')
    entry(proto_path, header_path, True)

    # Step 2: h -> out
    cpp_path = os.path.join(parent, '../cpp/')
    subprocess.call(['make', 'clean'], cwd=cpp_path)
    assert(subprocess.call(['make'], cwd=cpp_path) == 0)

    # Step 3: parse json
    exe_path = os.path.join(parent, '../cpp/main.out')
    json_path = os.path.join(parent, f'json/{name}.json')
    assert(subprocess.call([exe_path, json_path]) != 0)

    # clean
    subprocess.call(['make', 'clean'], cwd=cpp_path)
    os.remove(header_path)


def test_nothing():
    pipeline('nothing')


def test_invalid_empty():
    pipeline('invalid_empty')


def test_wrong_type():
    pipeline('wrong_type')


def test_missing_item():
    pipeline('missing_item')


def test_noarray():
    pipeline('noarray')

def test_eof():
    pipeline('eof')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='proto2rapidjson',
    version='1.1.4',
    packages=find_packages(),
    python_requires=">=3.6",
    description='Convert .proto file to header-only RapidJSON based c++ code',
    author='Jun Zhang',
    author_email='zhangjun990222@gmail.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Sweetnow/proto2rapidjson",
    entry_points={
        'console_scripts': [
            'command-name = proto2rapidjson.__main__:entry',
        ],
    },
    zip_safe=False
)
